@@ -29,6 +29,7 @@ struct station;
 struct network;
 struct scan_bss;
 struct handshake_state;
+struct erp_cache_entry;
 
 void network_connected(struct network *network);
 void network_disconnected(struct network *network);
@@ -84,10 +85,11 @@ struct l_dbus_message *network_connect_new_hidden_network(
 
 void network_blacklist_add(struct network *network, struct scan_bss *bss);
 
-const struct iovec *network_get_extra_ies(struct network *network,
-						size_t *num_elems);
-
-bool network_has_erp_identity(struct network *network);
+struct erp_cache_entry *network_get_erp_cache(struct network *network);
 
 const struct l_queue_entry *network_bss_list_get_entries(
 						struct network *network);
+
+struct l_dbus_message *__network_connect(struct network *network,
+						struct scan_bss *bss,
+						struct l_dbus_message *message);
