@@ -595,15 +595,8 @@ static int resolve_init(void)
 
 	method_name = l_settings_get_value(iwd_get_config(), "Network",
 						"NameResolvingService");
-	if (!method_name) {
-		method_name = l_settings_get_value(iwd_get_config(), "General",
-							"dns_resolve_method");
-		if (method_name)
-			l_warn("[General].dns_resolve_method is deprecated, "
-				"use [Network].NameResolvingService");
-		else /* Default to systemd-resolved service. */
-			method_name = "systemd";
-	}
+	if (!method_name)
+		method_name = "systemd";
 
 	for (i = 0; resolve_method_ops_list[i].name; i++) {
 		if (strcmp(resolve_method_ops_list[i].name, method_name))
