@@ -51,6 +51,7 @@ struct scan_bss {
 	int32_t signal_strength;
 	uint16_t capability;
 	uint8_t *rsne;
+	uint8_t *rsnxe;
 	uint8_t *wpa;
 	uint8_t *osen;
 	uint8_t *wsc;		/* Concatenated WSC IEs */
@@ -82,6 +83,9 @@ struct scan_bss {
 	bool vht_capable : 1;
 	bool anqp_capable : 1;
 	bool hs20_capable : 1;
+	bool force_default_sae_group : 1;
+	uint8_t cost_level : 3;
+	uint8_t cost_flags : 4;
 };
 
 struct scan_parameters {
